/*
 MIDI Synthesizer Demo for Synthino board
 Copyright (C) 2014 nootropic design, LLC
 All rights reserved.
 */

#include <Synthino.h>
#include <SoftwareSerial.h>
#include <MIDI.h>
#include <EEPROM.h>
#include <analogmuxdemux.h>
#include "waveforms.h"

//#define DEBUG_ENABLE
//#define TEST
#define BUTTON1_PIN 6
#define BUTTON2_PIN 7
#define BUTTON3_PIN 8
#define BUTTON4_PIN 9
#define LED 3


#define BUTTON1 0
#define BUTTON2 1
#define BUTTON3 2
#define BUTTON4 3


byte ATTACK_TIME_POT;
byte DECAY_TIME_POT;
byte SUSTAIN_LEVEL_POT;
byte RELEASE_TIME_POT;
byte LFO_FREQ_POT;
byte LFO_DEPTH_POT;
byte WAVEFORM_SELECT_POT;
byte PITCH_POT;

#define SILENCE 2048
#define MIDI_LOW 21
#define MIDI_HIGH 108
#define NOTE_OFF 0
#define NOTE_PENDING_OFF 1
#define MAX_NOTES 4
#define MAX_NOTE_VOL 32
#define DEFAULT_VELOCITY 40
#define MAX_VELOCITY 70
#define BASE_TUNING_ADJUSTMENT 1.0
#define EEPROM_MAGIC_NUMBER 0xbad0
#define POLYPHONIC_RATE 15000

#define ATTACK 0
#define DECAY 1
#define SUSTAIN 2
#define RELEASE 3

byte numNotes = 4;
boolean debug = false;
#ifdef DEBUG_ENABLE
//SoftwareSerial softSerial(A4, A5); // RX, TX
#endif

AnalogMux amux(2, 5, 4, 0);

byte masterWaveform;
byte noteWaveform[MAX_NOTES];
prog_int16_t *noteWaveformBuf[MAX_NOTES];
prog_int16_t *lfoWaveformBuf = waveformBuffers[0];
int tuningSetting = 0;

byte buttonPins[4] = {BUTTON1_PIN, BUTTON2_PIN, BUTTON3_PIN, BUTTON4_PIN};
byte buttonState[4] = {HIGH, HIGH, HIGH, HIGH};
byte buttonReleaseState[4] = {HIGH, HIGH, HIGH, HIGH};

unsigned int output = SILENCE;
unsigned int playbackSampleRate;
unsigned int timer1Start;
volatile unsigned int timer1End;

float noteTable[(MIDI_HIGH-MIDI_LOW)+1];

int noteVolLevelDuration[MAX_NOTES];
volatile int noteVolLevelRemaining[MAX_NOTES];
byte noteEnvelopePhase[MAX_NOTES];
byte noteVolIndex[MAX_NOTES];
byte noteTargetVolIndex[MAX_NOTES];
int noteVolIndexInc[MAX_NOTES];
int attackVolLevelDuration;
int decayVolLevelDuration;
float sustainVolLevel;
int releaseVolLevelDuration;
boolean isMIDINote[MAX_NOTES];

float lfoFrequency = 0.0;
float lfoDepth = 0.0;

byte buttonNote[MAX_NOTES]; // note currently assigned to button
byte lastButtonNote[MAX_NOTES]; // allows us to find the last note used when released.
int initialPitchReading[MAX_NOTES]; // pitch pot position when button pressed, expressed as MIDI note
byte initialWaveformReading[MAX_NOTES]; // waveform pot position when button pressed, expressed as MIDI note
boolean absolutePitch = false;

volatile byte notes[MAX_NOTES];
volatile float frequency[MAX_NOTES];
long noteStartTime[MAX_NOTES];
byte volume[MAX_NOTES];
byte volumeNext[MAX_NOTES];
volatile unsigned int phase[MAX_NOTES];
volatile unsigned int phaseInc[MAX_NOTES];
volatile byte phaseFraction[MAX_NOTES];
volatile byte phaseFractionInc[MAX_NOTES];

volatile unsigned int lfoPhase = 0;
volatile unsigned int lfoPhaseInc = 0;
volatile byte lfoPhaseFraction = 0;
volatile byte lfoPhaseFractionInc = 0;

float lfoShift;

unsigned long lastDebugPrint = 0;

void setup() {
#ifdef DEBUG_ENABLE
  Serial.begin(115200);
#endif

  playbackSampleRate = POLYPHONIC_RATE;
  timer1Start = UINT16_MAX - (F_CPU / playbackSampleRate);

  pinMode(BUTTON1_PIN, INPUT_PULLUP);
  pinMode(BUTTON2_PIN, INPUT_PULLUP);
  pinMode(BUTTON3_PIN, INPUT_PULLUP);
  pinMode(BUTTON4_PIN, INPUT_PULLUP);
  pinMode(LED, OUTPUT);


  masterWaveform = 0;
  for(int i=0;i<MAX_NOTES;i++) {
    notes[i] = NOTE_OFF;
    isMIDINote[i] = false;
    frequency[i] = 0;
    noteWaveform[i] = 0;
    noteWaveformBuf[i] = waveformBuffers[0];
    phase[i] = 0;
    phaseInc[i] = 0;
    phaseFraction[i] = 0;
    phaseFractionInc[i] = 0;
  }

  buttonNote[0] = 60;
  lastButtonNote[0] = 60;
  buttonNote[1] = 64;
  lastButtonNote[1] = 64;
  buttonNote[2] = 67;
  lastButtonNote[2] = 67;
  buttonNote[3] = 72;
  lastButtonNote[3] = 72;


  Synthino.begin();

  float tuningAdjustment = BASE_TUNING_ADJUSTMENT;
  readEEPROM();
  if (buttonPressed(BUTTON4)) {
    beep(50);
    delay(25);
    beep(50);
    delay(200);

    // start playing an A4
    initNote(0, 69);
    frequency[0] = 440.0;
    setPhaseIncrement(0);
    volume[0] = 6;
    volumeNext[0] = 6;

    while (buttonPressed(BUTTON4)) {
      tuningSetting = map(amux.AnalogRead(7), 0, 1023, -512, 511);
      tuningAdjustment = BASE_TUNING_ADJUSTMENT + ((float)tuningSetting) / 256.0; // [-2, 2]
      frequency[0] = 440.0 * (pow(2, (tuningAdjustment)/12.0));
#ifdef DEBUG_ENABLE
      delay(100);
      Serial.print("tuningAdjustment = ");
      Serial.print(tuningAdjustment);
      Serial.print("   f = ");
      Serial.println(frequency[0]);
#endif
      setPhaseIncrement(0);
      delay(10);
    }
    notes[0] = NOTE_PENDING_OFF;
    writeEEPROM();
  }

  ATTACK_TIME_POT = 0;
  DECAY_TIME_POT = 1;
  SUSTAIN_LEVEL_POT = 2;
  RELEASE_TIME_POT = 3;
  LFO_FREQ_POT = 4;
  LFO_DEPTH_POT = 5;
  WAVEFORM_SELECT_POT = 6;
  PITCH_POT = 7;
  timer1Start = UINT16_MAX - (F_CPU / playbackSampleRate);
    

  // precompute frequencies. Indexed by MIDI note number.
  tuningAdjustment = BASE_TUNING_ADJUSTMENT + ((float)tuningSetting) / 256.0; // [-2, 2]
  for(int i=MIDI_LOW;i<=MIDI_HIGH;i++) {
    noteTable[i-MIDI_LOW] = 440.0 * (pow(2, ((i-69)+tuningAdjustment)/12.0));
  }

  // Initialize MIDI library.
#ifndef TEST
  MIDI.begin(MIDI_CHANNEL_OMNI);    
  MIDI.setHandleNoteOn(handleNoteOn);
  MIDI.setHandleNoteOff(handleNoteOff);
#endif

#ifdef DEBUG_ENABLE
  Serial.print(" tuningAdjustment=");
  Serial.print(tuningAdjustment);
  Serial.print(" memory: ");
  Serial.print(getMemory());
  Serial.println();
#endif

}

void readEEPROM() {
  if (EEPROMValid()) {
    tuningSetting = EEPROM.read(3);
    tuningSetting = tuningSetting << 8;
    tuningSetting |= EEPROM.read(4);
#ifdef DEBUG_ENABLE
    Serial.print("  tuningSetting = ");
    Serial.println(tuningSetting);
#endif
  }
}

void writeEEPROM() {
  setEEPROMValid();
  EEPROM.write(3, tuningSetting >> 8);
  EEPROM.write(4, (tuningSetting & 0xFF));
}

boolean EEPROMValid() {
  // determine if the EEPROM has ever been written by this firmware
  // so we can determine if the values can be trusted
  unsigned int magic = EEPROM.read(0);
  magic = magic << 8;
  magic |= EEPROM.read(1);
#ifdef DEBUG_ENABLE
  if (magic == EEPROM_MAGIC_NUMBER) {
    Serial.println("EEPROM valid");
  } else {
    Serial.print("EEPROM NOT valid: ");
    Serial.println(magic, 16);
  }
#endif
  return (magic == EEPROM_MAGIC_NUMBER);
}
  
void setEEPROMValid() {
  EEPROM.write(0, EEPROM_MAGIC_NUMBER >> 8);
  EEPROM.write(1, (EEPROM_MAGIC_NUMBER & 0xFF));
}
  


void beep(int duration) {
  initNote(0, 81);
  frequency[0] = 880.0;
  setPhaseIncrement(0);
  volume[0] = 10;
  volumeNext[0] = 10;
  delay(duration);
  notes[0] = NOTE_PENDING_OFF;
}

void loop() {
  boolean lfo = false;

#ifdef DEBUG_ENABLE
  debug = buttonPressed(BUTTON4);
  doDebug();
#endif
  MIDI.read();

  boolean buttonNotesUsed = false;
  for(byte i=0;i<numNotes;i++) {
    if (buttonPressedNew(i)) {
      initialPitchReading[i] = map(amux.AnalogRead(PITCH_POT), 0, 1023, 36, 84);
      initialWaveformReading[i] = map(amux.AnalogRead(WAVEFORM_SELECT_POT), 0, 1023, 0, N_WAVEFORMS-1);
      if (notes[i] != NOTE_OFF) {
	// if the note is still being played (in RELEASE), turn it off.
	notes[i] = NOTE_PENDING_OFF;
	while (notes[i] != NOTE_OFF) {
	  // wait
	}
      }
      initNote(i, buttonNote[i]);
      noteWaveformBuf[i] = waveformBuffers[noteWaveform[i]];
      noteEnvelopePhase[i] = ATTACK;
      noteTargetVolIndex[i] = (byte)(((float)(DEFAULT_VELOCITY / 100.0)) * MAX_NOTE_VOL);
      noteVolIndexInc[i] = 1;
      noteVolLevelDuration[i] = attackVolLevelDuration;
      noteVolLevelRemaining[i] = attackVolLevelDuration;
      if (attackVolLevelDuration > 0) {
	noteVolIndex[i] = 0;
      } else {
	// optimize if there is no attack.  go directly to target volume
	noteVolIndex[i] = noteTargetVolIndex[i];
	volumeNext[i] = noteVolIndex[i];
      }
    } else {
      if ((notes[i] > NOTE_PENDING_OFF) && (!isMIDINote[i])) {
	if (buttonPressed(i)) {
	  buttonNotesUsed = true;
	  int currentPitchReading = map(amux.AnalogRead(PITCH_POT), 0, 1023, 36, 84);
	  if (currentPitchReading != initialPitchReading[i]) {
	    // the pot has moved since the button was pressed
	    if (absolutePitch) {
	      // pitch depends on absolute position of pot
	      buttonNote[i] = currentPitchReading;
	      initialPitchReading[i] = buttonNote[i];
	    } else {
	      // pitch depends on position of pot relative to when button pressed
	      buttonNote[i] += (currentPitchReading - initialPitchReading[i]);
	      initialPitchReading[i] = currentPitchReading;
	    }
	    if (buttonNote[i] < MIDI_LOW) {
	      buttonNote[i] = MIDI_LOW;
	    }
	    if (buttonNote[i] > MIDI_HIGH) {
	      buttonNote[i] = MIDI_HIGH;
	    }
	    notes[i] = buttonNote[i];
	  }
	  if (buttonNote[i] == lastButtonNote[i]) {
	    digitalWrite(LED, HIGH);
	  } else {
	    digitalWrite(LED, LOW);
	  }
	  byte currentWaveformReading = map(amux.AnalogRead(WAVEFORM_SELECT_POT), 0, 1023, 0, N_WAVEFORMS-1);
	  if (currentWaveformReading != initialWaveformReading[i]) {
	    // the pot has moved since the button was pressed
	    noteWaveform[i] = currentWaveformReading;
	    noteWaveformBuf[i] = waveformBuffers[noteWaveform[i]];
	    initialWaveformReading[i] = currentWaveformReading;
	  }
	} else {
	  if (noteEnvelopePhase[i] != RELEASE) {
	    // button released
	    lastButtonNote[i] = notes[i];
	    digitalWrite(A5, LOW);
	    noteEnvelopePhase[i] = RELEASE;
	    noteTargetVolIndex[i] = 0;
	    noteVolIndexInc[i] = -1;
	    noteVolLevelDuration[i] = releaseVolLevelDuration;
	    noteVolLevelRemaining[i] = releaseVolLevelDuration;
	    if (releaseVolLevelDuration == 0) {
	      // optimize for immediate release
	      noteVolIndex[i] = 0;
	      volumeNext[i] = 0;
	    }
	  }
	}
      }
    }
  }

  byte lastMasterWaveform = masterWaveform;
  masterWaveform = map(amux.AnalogRead(WAVEFORM_SELECT_POT), 0, 1023, 0, N_WAVEFORMS-1);
  if (masterWaveform != lastMasterWaveform) {
    for(byte i=0;i<numNotes;i++) {
      noteWaveformBuf[i] = waveformBuffers[masterWaveform];
    }
  }

  int lfoFreqReading = amux.AnalogRead(LFO_FREQ_POT);
  if (lfoFreqReading > 0) {
    lfo = true;
    lfoFrequency = lfoFreqReading / 50.0; // 0.0-20.0
    lfoDepth = amux.AnalogRead(LFO_DEPTH_POT) / 2048.0;  // 0.0-0.5

    float lfoPhaseIncFloat = (lfoFrequency * N_WAVEFORM_SAMPLES) / playbackSampleRate;
    lfoPhaseInc = (int)lfoPhaseIncFloat;
    lfoPhaseIncFloat -= lfoPhaseInc;
    lfoPhaseFractionInc = (int)(lfoPhaseIncFloat * 256.0);

    // lfoShift is how much to change the frequency.  Range is [-1.0:1.0]
    lfoShift = (((int)pgm_read_word(lfoWaveformBuf + lfoPhase) - 2048)/2048.0);
  }

  attackVolLevelDuration = amux.AnalogRead(ATTACK_TIME_POT) / 2;
  decayVolLevelDuration = amux.AnalogRead(DECAY_TIME_POT) / 2;
  sustainVolLevel = amux.AnalogRead(SUSTAIN_LEVEL_POT) / 1023.0;
  releaseVolLevelDuration = amux.AnalogRead(RELEASE_TIME_POT) / 2;

  boolean notePlaying = false;
  for(byte i=0;i<numNotes;i++) {
    if (notes[i] > NOTE_PENDING_OFF) {
      if ((notes[i] < MIDI_LOW) || (notes[i] > MIDI_HIGH)) {
	continue;
      }
      notePlaying = true;
      frequency[i] = noteTable[notes[i]-MIDI_LOW];

      processEnvelope(i);

      // Adjust each note frequency for the LFO effect.
      if (lfo) {
        frequency[i] += frequency[i] * (lfoShift * lfoDepth);
      }

      setPhaseIncrement(i);

    }
  }

  if (!buttonNotesUsed) {
    if (notePlaying) {
      digitalWrite(LED, HIGH);
    } else {
      digitalWrite(LED, LOW);
    }
  }

}

void setPhaseIncrement(byte noteIndex) {
  // phaseInc is the number of steps we should progress through the
  // wavetable at each execution of the ISR.
  float phaseIncFloat = (frequency[noteIndex] * N_WAVEFORM_SAMPLES) / playbackSampleRate;
  // convert the phase increment to an int
  phaseInc[noteIndex] = (int)phaseIncFloat;
  // but also use a counter to keep track of the fractional part
  phaseIncFloat -= phaseInc[noteIndex];
  phaseFractionInc[noteIndex] = (byte)(phaseIncFloat * 256.0);
}


void processEnvelope(byte i) {
  if (noteEnvelopePhase[i] == SUSTAIN) {
    return;
  }
  if (noteVolLevelRemaining[i] <= 0) {
    // it is time to change the volume to the next level
    if (noteVolIndex[i] == noteTargetVolIndex[i]) {
      // this was the last volume level to use for this phase.
      // switch to new phase of envelope.
      byte phase = noteEnvelopePhase[i];
      if (phase == ATTACK) {
        noteEnvelopePhase[i] = DECAY;
        // decay to the sustain volume which is a percentage of the original velocity
        noteTargetVolIndex[i] = (byte)(sustainVolLevel * noteTargetVolIndex[i]);
        noteVolIndexInc[i] = -1; // decreasing volume
        noteVolLevelDuration[i] = decayVolLevelDuration;
        noteVolLevelRemaining[i] = decayVolLevelDuration;
      } 
      else {
        if (phase == DECAY) {
          noteEnvelopePhase[i] = SUSTAIN;
        } 
        else {
          if (phase == RELEASE) {
            // done with this note.
            notes[i] = NOTE_PENDING_OFF;
          }
        }
      }
    } 
    else {
      // not yet at the target volume.
      // Take one step closer by changing the volume.
      noteVolIndex[i] += noteVolIndexInc[i];
      noteVolLevelRemaining[i] = noteVolLevelDuration[i];
      volumeNext[i] = noteVolIndex[i];
    }
  }
}

void initNote(byte index, byte note) {
  isMIDINote[index] = false;
  volume[index] = 0;
  volumeNext[index] = 0;
  phase[index] = 0;
  phaseInc[index] = 0;
  phaseFraction[index] = 0;
  phaseFractionInc[index] = 0;
  notes[index] = note;
}

// find a note in the array of MAX_NOTES available notes to use
byte findNoteIndex() {
  for(byte i=0;i<numNotes;i++) {
    // find a slot for this note
    if (notes[i] <= NOTE_PENDING_OFF) {
      while (notes[i] != NOTE_OFF) {
	// wait until note reaches NOTE_OFF state to avoid click
      }
      return i;
    }
  }
  // there are no notes that are not being used. Check to see if any
  // are in RELEASE envelope phase and use it
  for(byte i=0;i<numNotes;i++) {
    if (noteEnvelopePhase[i] == RELEASE) {
      notes[i] = NOTE_PENDING_OFF;
      while (notes[i] != NOTE_OFF) {
	// wait until note reaches NOTE_OFF state to avoid click
      }
      return i;
    }
  }
  // All notes are being used. Choose the note that is "oldest".
  byte noteIndex = 0;
  for(byte i=1;i<numNotes;i++) {
    if (noteStartTime[i] < noteStartTime[noteIndex]) {
      noteIndex = i;
    }
  }
  notes[noteIndex] = NOTE_PENDING_OFF;
  while (notes[noteIndex] != NOTE_OFF) {
    // wait until note reaches NOTE_OFF state to avoid click
  }
  return noteIndex;
}

void handleNoteOn(byte channel, byte note, byte velocity) { 
#ifdef DEBUG_ENABLE
  if (debug) {
    Serial.print("ON ");
    Serial.print(note);
    Serial.print(" : ");
    Serial.println(velocity);
  }
#endif
  if (velocity > 0) {
    byte noteIndex = findNoteIndex();
    if (velocity > MAX_VELOCITY) {
      velocity = MAX_VELOCITY;
    }
    initNote(noteIndex, note);
    noteStartTime[noteIndex] = millis();
    noteWaveformBuf[noteIndex] = waveformBuffers[masterWaveform];
    isMIDINote[noteIndex] = true;
    noteEnvelopePhase[noteIndex] = ATTACK;
    noteTargetVolIndex[noteIndex] = (byte)(((float)(velocity / 100.0)) * MAX_NOTE_VOL);
    noteVolIndexInc[noteIndex] = 1;
    noteVolLevelDuration[noteIndex] = attackVolLevelDuration;
    noteVolLevelRemaining[noteIndex] = attackVolLevelDuration;
    if (attackVolLevelDuration > 0) {
      noteVolIndex[noteIndex] = 0;
    } else {
      // optimize if there is no attack.  go directly to target volume
      noteVolIndex[noteIndex] = noteTargetVolIndex[noteIndex];
      volumeNext[noteIndex] = noteVolIndex[noteIndex];
    }
  } else {
    // note off
    for(byte i=0;i<numNotes;i++) {
      // find the note
      if (notes[i] == note) {
        noteEnvelopePhase[i] = RELEASE;
        noteTargetVolIndex[i] = 0;
        noteVolIndexInc[i] = -1;
        noteVolLevelDuration[i] = releaseVolLevelDuration;
        noteVolLevelRemaining[i] = releaseVolLevelDuration;
	if (releaseVolLevelDuration == 0) {
	  // optimize for immediate release
	  noteVolIndex[i] = 0;
	  volumeNext[i] = 0;
	}
      }
    }
  }
}

void handleNoteOff(byte channel, byte note, byte velocity) { 
#ifdef DEBUG_ENABLE
  if (debug) {
    Serial.print("OFF ");
    Serial.print(note);
    Serial.print(" : ");
    Serial.println(velocity);
  }
#endif
  for(byte i=0;i<numNotes;i++) {
    // find the note
    if (notes[i] == note) {
      noteEnvelopePhase[i] = RELEASE;
      noteTargetVolIndex[i] = 0;
      noteVolIndexInc[i] = -1;
      noteVolLevelDuration[i] = releaseVolLevelDuration;
      noteVolLevelRemaining[i] = releaseVolLevelDuration;
      //break; // bug fix for original InSoc code. Release all notes, not just first one.
    }
  }
}


ISR(TIMER1_OVF_vect) {
  TCNT1 = timer1Start;
  byte last;

  Synthino.output(output);

  lfoPhase += lfoPhaseInc;
  last = lfoPhaseFraction;
  lfoPhaseFraction += lfoPhaseFractionInc;
  if (lfoPhaseFraction < last) {
    // overflow in pseudo-floating point counter
    lfoPhase++;
  }
  if (lfoPhase >= N_WAVEFORM_SAMPLES) {
    lfoPhase -= N_WAVEFORM_SAMPLES;
  }


  output = SILENCE;


  //===============
  int mix;
  int sum = 0;
  for(byte i=0;i<numNotes;i++) {
    if (notes[i] != NOTE_OFF) {
      phase[i] += phaseInc[i];
      last = phaseFraction[i];
      phaseFraction[i] += phaseFractionInc[i];
      if (phaseFraction[i] < last) {
        // overflow in pseudo-floating point counter
        phase[i]++;
      }
      if (phase[i] >= N_WAVEFORM_SAMPLES) {
        if (notes[i] != NOTE_PENDING_OFF) {
          phase[i] -= N_WAVEFORM_SAMPLES;
          // don't assign volume until we cross zero
          volume[i] = volumeNext[i];
        } 
        else {
          notes[i] = NOTE_OFF;
          continue;
        }
      }
      noteVolLevelRemaining[i]--;
      mix = pgm_read_word(noteWaveformBuf[i] + phase[i]);
      mix = adjustAmplitude(mix, volume[i]);
      sum += mix;
    }
  }

  // clip
  if (sum < -2048) {
    sum = -2048;
  } 
  else {
    if (sum > 2047) {
      sum = 2047;
    }
  }

  output = sum + 2048;

#ifdef DEBUG_ENABLE
  timer1End = TCNT1;
#endif
}


/*
 * Adjust amplitude using only bitshifting.  Crude but cheap.
 */
int adjustAmplitude(int v, byte level) {
  int tmp, tmp2;
  switch(level) {
  case 32:
    return v;
    break;
  case 31:
    return v - (v >> 5);
    break;
  case 30:
    return v - (v >> 4);
    break;
  case 29:
    tmp = (v >> 4);
    return (v - tmp) - (tmp >> 1);
    break;
  case 28:
    return v - (v >> 3);
    break;
  case 27:
    tmp = v >> 3;
    return (v - tmp) - (tmp >> 2);
    break;
  case 26:
    tmp = v >> 3;
    return (v - tmp) - (tmp >> 1);
    break;
  case 25:
    tmp = v >> 2;
    return (v - tmp) + (tmp >> 3);
    break;
  case 24:
    return v - (v >> 2);
    break;
  case 23:
    tmp = v >> 2;
    return (v - tmp) - (tmp >> 3);
    break;
  case 22:
    tmp = v >> 2;
    return (v - tmp) - (tmp >> 2);
    break;
  case 21:
    tmp = v >> 1;
    tmp2 = tmp >> 2;
    return tmp + tmp2 + (tmp2 >> 2); // bad
    break;
  case 20: 
    tmp = v >> 1;
    return tmp + (tmp >> 2);
    break;
  case 19:
    tmp = v >> 1;
    tmp2 = tmp >> 3;
    return tmp + tmp2 + (tmp2 >> 2); // bad
    break;
  case 18:
    tmp = v >> 1;
    return tmp + (tmp >> 3);
    break;
  case 17:
    tmp = v >> 1;
    return tmp + (tmp >> 4);
    break;
  case 16:
    return v >> 1;
    break;
  case 15:
    tmp = v >> 1;
    return tmp - (tmp >> 4);
    break;
  case 14:
    tmp = v >> 1;
    return tmp - (tmp >> 3);
    break;
  case 13:
    tmp = v >> 1;
    tmp2 = tmp >> 3;
    return (tmp - tmp2) - (tmp2 >> 1); // bad
    break;
  case 12:
    tmp = v >> 1;
    return tmp - (tmp >> 2);
    break;
  case 11:
    tmp = v >> 1;
    tmp2 = tmp >> 2;
    return (tmp - tmp2) - (tmp2 >> 2);
    break;
  case 10:
    tmp = v >> 2;
    return tmp + (tmp >> 2);
    break;
  case 9:
    tmp = v >> 2;
    return tmp + (tmp >> 3);
    break;
  case 8:
    return v >> 2;
    break;
  case 7:
    tmp = v >> 2;
    return tmp - (tmp >> 3);
    break;
  case 6:
    tmp = v >> 3;
    return tmp + (tmp >> 1);
    break;
  case 5:
    tmp = v >> 3;
    return tmp + (tmp >> 2);
    break;
  case 4:
    return v >> 3;
  case 3:
    tmp = v >> 4;
    return tmp + (tmp >> 1);
    break;
  case 2:
    return v >> 4;
  case 1:
    return v >> 5;
  default:
    return 0;
  }
}


boolean buttonPressed(byte button) {
  return buttonPressed(button, true);
}

// Return true if the button is pressed.
boolean buttonPressed(byte button, boolean updateButtonState) {
  if (digitalRead(buttonPins[button]) == LOW) {
    // the button is currently pressed
    if ((buttonState[button] == HIGH) && (updateButtonState)) {
      // if the button was not pressed before, update the state.
      buttonState[button] = LOW;
    }
    return true;
  } else {
    // The button is currently not pressed
    if ((buttonState[button] == LOW) && (updateButtonState)) {
      // if the button was pressed before, update the state.
      buttonState[button] = HIGH;
    }
    return false;
  }
}

// Return true if the button is pressed and it is a new press (not held)
boolean buttonPressedNew(byte button) {
  if (digitalRead(buttonPins[button]) == LOW) {
    // The button is currently pressed
    if (buttonState[button] == HIGH) {
      // This is a new press.
      buttonState[button] = LOW;
      return true;
    }
    // This is not a new press.
    return false; 
  } 
  else {
    // The button is currently not pressed
    if (buttonState[button] == LOW) {
      buttonState[button] = HIGH;
    }
    return false;
  }
}

// Return true if the button is not pressed and it has been released since last check.
boolean buttonReleased(byte button) {
  if (digitalRead(buttonPins[button]) == HIGH) {
    // The button is currently not pressed
    if (buttonReleaseState[button] == LOW) {
      // This is a new release.
      buttonReleaseState[button] = HIGH;
      return true;
    }
    // This is not a new release
    return false; 
  } 
  else {
    // The button is currently pressed
    if (buttonReleaseState[button] == HIGH) {
      buttonReleaseState[button] = LOW;
    }
    return false;
  }
}


#ifdef DEBUG_ENABLE
void doDebug() {
  if (debug) {
    if ((millis() - lastDebugPrint) >= 400) {
      lastDebugPrint = millis();

      // Print the number of instruction cycles remaining at the end of the ISR.
      // The more work you try to do in the ISR, the lower this number will become.
      // If the number of cycles remaining reaches 0, then the ISR will take up
      // all the CPU time and the code in v will not run.

      Serial.print("mem: ");
      Serial.print(getMemory());
      Serial.print(" ");

      /*
      Serial.print(" rate: ");
      Serial.print(playbackSampleRate);
      Serial.print(" ");

      Serial.print("attack: ");
      Serial.print(attackVolLevelDuration);
      Serial.print(", decay: ");
      Serial.print(decayVolLevelDuration);
      Serial.print(", sustain: ");
      Serial.print(sustainVolLevel);
      Serial.print(", release: ");
      Serial.print(releaseVolLevelDuration);
      Serial.println("");
      */
      for(byte i=0;i<numNotes;i++) {
        Serial.print("notes[");
        Serial.print(i);
        Serial.print("]=");
        Serial.print(notes[i]);
        Serial.print("  ");

        Serial.print("f[");
        Serial.print(i);
        Serial.print("]=");
        Serial.print(frequency[i]);

	/*
        Serial.print(" pi[");
        Serial.print(i);
        Serial.print("]=");
        Serial.print(phaseInc[i]);
	*/

        Serial.print(" vol[");
        Serial.print(i);
        Serial.print("]=");
        Serial.print(volume[i]);
        Serial.print(" | ");
      }

      /*
      Serial.print("lfoFrequency = ");
      Serial.print(lfoFrequency, 4);
      Serial.print("  ");
      */

      Serial.print(" cycles = ");
      Serial.print(UINT16_MAX - timer1End);
      Serial.println("");
      if ((UINT16_MAX - timer1End) < 20) {
        Serial.println("WARNING: ISR execution time is too long. Reduce sample rate or reduce the amount of code in the ISR.");
      }
    }
  } // if (debug)
}


int getMemory() {
  int size = 200;
  byte *buf;
  while ((size > 0) && ((buf = (byte *) malloc(--size)) == NULL));
  free(buf);
  return size;
} 

#endif

