/*
 MIDI Synthesizer Demo for Synthino board
 Copyright (C) 2014 nootropic design, LLC
 All rights reserved.
 */

#include <Synthino.h>
#include <MIDI.h>
#include <EEPROM.h>
#include <analogmuxdemux.h>
#include "SynthinoDemo.h"
#include "waveforms.h"
#ifdef DEBUG_ENABLE
#include <SoftwareSerial.h>
#endif

byte ATTACK_TIME_POT;
byte DECAY_TIME_POT;
byte SUSTAIN_LEVEL_POT;
byte RELEASE_TIME_POT;
byte LFO_WAVEFORM_SELECT_POT;
byte LFO_FREQ_POT;
byte LFO_DEPTH_POT;
byte WAVEFORM_SELECT_POT;
byte PITCH_POT;

// overall mode of the firmware.
// mode 1 is default 4-voice polyphonic synth
// mode 2 = monosynth at MONOSYNTH_RATE
// mode 3 = AR envelope on pots 2,3.  Pot 1 controls LFO waveform
byte mode = 1;
byte numNotes = 4;
note_t *note;
channel_t *channel;
button_t *button;
boolean debug = false;
#ifdef DEBUG_ENABLE
SoftwareSerial softSerial(A4, A5); // RX, TX
#endif

AnalogMux amux(2, 5, 4, 0);

byte masterWaveform;
prog_int16_t *lfoWaveformBuf = waveformBuffers[0];
int tuningSetting = 0;

byte buttonPins[4] = {BUTTON1_PIN, BUTTON2_PIN, BUTTON3_PIN, BUTTON4_PIN};
byte buttonState[4] = {HIGH, HIGH, HIGH, HIGH};
byte buttonReleaseState[4] = {HIGH, HIGH, HIGH, HIGH};
byte ledState = LOW;

unsigned int output = SILENCE;
unsigned int playbackSampleRate;
float timeScale;
unsigned int timer1Start;
volatile unsigned int timer1End;

float noteTable[(MIDI_HIGH-MIDI_LOW)+1];


float lfoFrequency = 0.0;
float lfoDepth = 0.0;

byte selectedChannel = 1; // MIDI channel associated with the controls
boolean absolutePitch = false;

volatile unsigned int lfoPhase = 0;
volatile unsigned int lfoPhaseInc = 0;
volatile byte lfoPhaseFraction = 0;
volatile byte lfoPhaseFractionInc = 0;

float lfoShift;

unsigned long lastDebugPrint = 0;

void setup() {
#ifdef DEBUG_ENABLE
  softSerial.begin(115200);
#endif

  note = (note_t *)malloc(numNotes * sizeof(note_t));
  channel = (channel_t *)malloc(N_MIDI_CHANNELS * sizeof(channel_t));
  button = (button_t *)malloc(N_BUTTONS * sizeof(button_t));

  playbackSampleRate = POLYPHONIC_RATE;
  timer1Start = UINT16_MAX - (F_CPU / playbackSampleRate);
  timeScale = MONOSYNTH_RATE / playbackSampleRate;

  pinMode(BUTTON1_PIN, INPUT_PULLUP);
  pinMode(BUTTON2_PIN, INPUT_PULLUP);
  pinMode(BUTTON3_PIN, INPUT_PULLUP);
  pinMode(BUTTON4_PIN, INPUT_PULLUP);
  pinMode(LED, OUTPUT);


  masterWaveform = 0;
  for(int i=0;i<MAX_NOTES;i++) {
    note[i].midiVal = NOTE_OFF;
    note[i].midiChannel = 0; // zero means no MIDI channel
    note[i].frequency = 0;
    note[i].waveformBuf = waveformBuffers[0];
    note[i].phase = 0;
    note[i].phaseInc = 0;
    note[i].phaseFraction = 0;
    note[i].phaseFractionInc = 0;
  }


  Synthino.begin();


  float tuningAdjustment = BASE_TUNING_ADJUSTMENT;
  readEEPROM();
  if (buttonPressed(BUTTON4)) {
    beep(50);
    delay(25);
    beep(50);
    delay(200);

    // start playing an A4
    initNote(0, 69);
    note[0].frequency = 440.0;
    setPhaseIncrement(0);
    note[0].volume = 6;
    note[0].volumeNext = 6;
    while (buttonPressed(BUTTON4)) {
      tuningSetting = map(amux.AnalogRead(7), 0, 1023, -512, 511);
      tuningAdjustment = BASE_TUNING_ADJUSTMENT + ((float)tuningSetting) / 256.0; // [-2, 2]
      note[0].frequency = 440.0 * (pow(2, (tuningAdjustment)/12.0));
#ifdef DEBUG_ENABLE
      delay(100);
      softSerial.print("tuningAdjustment = ");
      softSerial.print(tuningAdjustment);
      softSerial.print("   f = ");
      softSerial.println(note[0].frequency);
#endif
      setPhaseIncrement(0);
      delay(10);
    }
    note[0].midiVal = NOTE_PENDING_OFF;
    writeEEPROM();
  }

  // Setting the mode
  if (buttonPressed(BUTTON1)) {
    mode = 255;
    while (buttonPressed(BUTTON1)) {
      delay(100);
      byte lastMode = mode;
      int modeReading = amux.AnalogRead(4);
      mode = map(amux.AnalogRead(4), 0, 1024, 1, 4);
#ifdef DEBUG_ENABLE
      softSerial.print("   r = ");
      softSerial.print(modeReading);
      softSerial.print("   mode = ");
      softSerial.println(mode);
#endif
      if (mode != lastMode) {
	for(byte i=0;i<mode;i++) {
	  beep(75);
	  delay(50);
	}
      }	
    }
    writeEEPROM();
  }
  if (mode == 1) {
    numNotes = 4;
    playbackSampleRate = POLYPHONIC_RATE;
    ATTACK_TIME_POT = 0;
    DECAY_TIME_POT = 1;
    SUSTAIN_LEVEL_POT = 2;
    RELEASE_TIME_POT = 3;
    LFO_FREQ_POT = 4;
    LFO_DEPTH_POT = 5;
    WAVEFORM_SELECT_POT = 6;
    PITCH_POT = 7;
  }
  if (mode == 2) {
    numNotes = 1;
    playbackSampleRate = MONOSYNTH_RATE;
    ATTACK_TIME_POT = 0;
    DECAY_TIME_POT = 1;
    SUSTAIN_LEVEL_POT = 2;
    RELEASE_TIME_POT = 3;
    LFO_FREQ_POT = 4;
    LFO_DEPTH_POT = 5;
    WAVEFORM_SELECT_POT = 6;
    PITCH_POT = 7;
  }
  if (mode == 3) {
    numNotes = 4;
    playbackSampleRate = POLYPHONIC_RATE;
    ATTACK_TIME_POT = 2;
    DECAY_TIME_POT = 3;
    RELEASE_TIME_POT = 3;
    LFO_WAVEFORM_SELECT_POT = 0;
    LFO_FREQ_POT = 4;
    LFO_DEPTH_POT = 5;
    WAVEFORM_SELECT_POT = 6;
    PITCH_POT = 7;
  }
  timeScale = (float)MONOSYNTH_RATE / (float)playbackSampleRate;
  timer1Start = UINT16_MAX - (F_CPU / playbackSampleRate);

  for(byte i=0;i<numNotes;i++) {
    button[i].pitchReading = map(amux.AnalogRead(PITCH_POT), 0, 1023, 36, 84);
    button[i].waveformReading = map(amux.AnalogRead(WAVEFORM_SELECT_POT), 0, 1024, 0, N_WAVEFORMS);
    button[i].attackReading = amux.AnalogRead(ATTACK_TIME_POT) / timeScale;
    button[i].decayReading = amux.AnalogRead(DECAY_TIME_POT) / timeScale;
    button[i].sustainReading = amux.AnalogRead(SUSTAIN_LEVEL_POT) / 1023.0;
    button[i].releaseReading = amux.AnalogRead(RELEASE_TIME_POT) / timeScale;
    channel[i].waveform = 0;
    channel[i].attackVolLevelDuration = button[i].attackReading;
    channel[i].decayVolLevelDuration = button[i].decayReading;
    channel[i].sustainVolLevel = button[i].sustainReading;
    channel[i].releaseVolLevelDuration = button[i].releaseReading;
  }

  button[0].note = 60;
  button[0].lastNote = 60;
  button[1].note = 64;
  button[1].lastNote = 64;
  button[2].note = 67;
  button[2].lastNote = 67;
  button[3].note = 72;
  button[3].lastNote = 72;

  // precompute frequencies. Indexed by MIDI note number.
  tuningAdjustment = BASE_TUNING_ADJUSTMENT + ((float)tuningSetting) / 256.0; // [-2, 2]
  for(int i=MIDI_LOW;i<=MIDI_HIGH;i++) {
    noteTable[i-MIDI_LOW] = 440.0 * (pow(2, ((i-69)+tuningAdjustment)/12.0));
  }

  // Initialize MIDI library.
#ifndef TEST
  MIDI.begin(MIDI_CHANNEL_OMNI);    
  MIDI.setHandleNoteOn(handleNoteOn);
  MIDI.setHandleNoteOff(handleNoteOff);
#endif

  digitalWrite(LED, HIGH);
  delay(50);
  digitalWrite(LED, LOW);

#ifdef DEBUG_ENABLE
  softSerial.print("setup complete: mode=");
  softSerial.print(mode);
  softSerial.print(" timeScale=");
  softSerial.print(timeScale);
  softSerial.print(" tuningAdjustment=");
  softSerial.print(tuningAdjustment);
  softSerial.print(" memory: ");
  softSerial.print(getMemory());
  softSerial.println();
#endif

}

void readEEPROM() {
  if (EEPROMValid()) {
    mode = EEPROM.read(2);
    tuningSetting = EEPROM.read(3);
    tuningSetting = tuningSetting << 8;
    tuningSetting |= EEPROM.read(4);
#ifdef DEBUG_ENABLE
    softSerial.print("mode = ");
    softSerial.print(mode);
    softSerial.print("  tuningSetting = ");
    softSerial.println(tuningSetting);
#endif
  }
}

void writeEEPROM() {
  setEEPROMValid();
  EEPROM.write(2, mode);
  EEPROM.write(3, tuningSetting >> 8);
  EEPROM.write(4, (tuningSetting & 0xFF));
}

boolean EEPROMValid() {
  // determine if the EEPROM has ever been written by this firmware
  // so we can determine if the values can be trusted
  unsigned int magic = EEPROM.read(0);
  magic = magic << 8;
  magic |= EEPROM.read(1);
#ifdef DEBUG_ENABLE
  if (magic == EEPROM_MAGIC_NUMBER) {
    softSerial.println("EEPROM valid");
  } else {
    softSerial.print("EEPROM NOT valid: ");
    softSerial.println(magic, 16);
  }
#endif
  return (magic == EEPROM_MAGIC_NUMBER);
}
  
void setEEPROMValid() {
  EEPROM.write(0, EEPROM_MAGIC_NUMBER >> 8);
  EEPROM.write(1, (EEPROM_MAGIC_NUMBER & 0xFF));
}
  


void beep(int duration) {
  note_t *beepNote = &note[0];
  initNote(0, 81);
  beepNote->frequency = 880.0;
  setPhaseIncrement(0);
  beepNote->volume = 10;
  beepNote->volumeNext = 10;
  delay(duration);
  beepNote->midiVal = NOTE_PENDING_OFF;
}

void loop() {
  boolean lfo = false;
  ledState = LOW;
  int diff;

#ifdef DEBUG_ENABLE
  debug = true;
  doDebug();
#endif
  MIDI.read();

  readButtons();

  // Compare the waveform and envelope pot position to the position they
  // had when the channel was selected. Only change the channel waveform 
  // and envelope seetings if the pots have moved.
  byte currentWaveformReading = map(amux.AnalogRead(WAVEFORM_SELECT_POT), 0, 1024, 0, N_WAVEFORMS);
  if (currentWaveformReading != button[selectedChannel-1].waveformReading) {
    // The waveform knob has moved since the channel was chosen.
    channel[selectedChannel-1].waveform = currentWaveformReading;
    for(byte i=0;i<numNotes;i++) {
      if (note[i].midiChannel == selectedChannel) {
	note[i].waveformBuf = waveformBuffers[currentWaveformReading];
      }
    }
    if (buttonHeld(selectedChannel-1)) {
      button[selectedChannel-1].waveformReading = currentWaveformReading;
      note[selectedChannel-1].waveformBuf = waveformBuffers[currentWaveformReading];
    }
  }

  int currentAttackReading = amux.AnalogRead(ATTACK_TIME_POT) / timeScale;
  diff = abs(currentAttackReading - button[selectedChannel-1].attackReading);
  if ((diff > POT_TOLERANCE) || ((diff > 0) && (currentAttackReading == 0))) {
    channel[selectedChannel-1].attackVolLevelDuration = currentAttackReading;
    if (buttonHeld(selectedChannel-1)) {
      button[selectedChannel-1].attackReading = currentAttackReading;
    }
  }

  int currentDecayReading = amux.AnalogRead(DECAY_TIME_POT) / timeScale;
  if (abs(currentDecayReading - button[selectedChannel-1].decayReading) > POT_TOLERANCE) {
    channel[selectedChannel-1].decayVolLevelDuration = currentDecayReading;
    if (buttonHeld(selectedChannel-1)) {
      button[selectedChannel-1].decayReading = currentDecayReading;
    }
  }

  float currentSustainReading = ((float)amux.AnalogRead(SUSTAIN_LEVEL_POT)) / 1023.0;
  float fdiff = currentSustainReading - button[selectedChannel-1].sustainReading;
  if ((fdiff > 0.10) || (fdiff < -0.10)) {
    if (mode != 3) {
      channel[selectedChannel-1].sustainVolLevel = currentSustainReading;
      if (buttonHeld(selectedChannel-1)) {
	button[selectedChannel-1].sustainReading = currentSustainReading;
      }
    } else {
      channel[selectedChannel-1].sustainVolLevel = 1.0;
    }
  }

  int currentReleaseReading = amux.AnalogRead(RELEASE_TIME_POT) / timeScale;
  diff = abs(currentReleaseReading - button[selectedChannel-1].releaseReading);
  if ((diff > POT_TOLERANCE) || ((diff > 0) && (currentReleaseReading == 0))) {
    channel[selectedChannel-1].releaseVolLevelDuration = currentReleaseReading;
    if (buttonHeld(selectedChannel-1)) {
      button[selectedChannel-1].releaseReading = currentReleaseReading;
    }
  }


  int lfoFreqReading = amux.AnalogRead(LFO_FREQ_POT);
  if (lfoFreqReading > 0) {
    lfo = true;
    lfoFrequency = lfoFreqReading / 50.0; // 0.0-20.0
    lfoDepth = amux.AnalogRead(LFO_DEPTH_POT) / 2048.0;  // 0.0-0.5

    float lfoPhaseIncFloat = (lfoFrequency * N_WAVEFORM_SAMPLES) / playbackSampleRate;
    lfoPhaseInc = (int)lfoPhaseIncFloat;
    lfoPhaseIncFloat -= lfoPhaseInc;
    lfoPhaseFractionInc = (int)(lfoPhaseIncFloat * 256.0);

    // lfoShift is how much to change the frequency.  Range is [-1.0:1.0]
    if (mode == 3) {
      byte lfoWaveform = map(amux.AnalogRead(LFO_WAVEFORM_SELECT_POT), 0, 1024, 0, N_LFO_WAVEFORMS);
      lfoWaveformBuf = lfoWaveformBuffers[lfoWaveform];
    }
    lfoShift = (((int)pgm_read_word(lfoWaveformBuf + lfoPhase) - 2048)/2048.0);
  }


  boolean midiNotePlaying = false;
  for(byte i=0;i<numNotes;i++) {
    if (note[i].midiVal > NOTE_PENDING_OFF) {
      if (note[i].midiChannel > 0) {
	ledState = HIGH;
      }
      note[i].frequency = noteTable[note[i].midiVal-MIDI_LOW];

      processEnvelope(i);

      // Adjust each note frequency for the LFO effect.
      if (lfo) {
        note[i].frequency += note[i].frequency * (lfoShift * lfoDepth);
      }

      setPhaseIncrement(i);

    }
  }

  digitalWrite(LED, ledState);

}

void readButtons() {
  for(byte i=0;i<numNotes;i++) {
    if (buttonPressedNew(i)) {
      selectedChannel = i+1;
      button[i].pitchReading = map(amux.AnalogRead(PITCH_POT), 0, 1023, 36, 84);
      button[i].waveformReading = map(amux.AnalogRead(WAVEFORM_SELECT_POT), 0, 1024, 0, N_WAVEFORMS);
      button[i].attackReading = amux.AnalogRead(ATTACK_TIME_POT) / timeScale;
      button[i].decayReading = amux.AnalogRead(DECAY_TIME_POT) / timeScale;
      button[i].sustainReading = amux.AnalogRead(SUSTAIN_LEVEL_POT) / 1023.0;
      button[i].releaseReading = amux.AnalogRead(RELEASE_TIME_POT) / timeScale;

      if (note[i].midiVal != NOTE_OFF) {
	// if the note is still being played (in RELEASE), turn it off.
	note[i].midiVal = NOTE_PENDING_OFF;
	while (note[i].midiVal != NOTE_OFF) {
	  // wait
	}
      }
      initNote(i, button[i].note);
      note[i].waveformBuf = waveformBuffers[channel[selectedChannel-1].waveform];
      note[i].envelopePhase = ATTACK;
      note[i].targetVolIndex = (byte)(((float)(DEFAULT_VELOCITY / 100.0)) * MAX_NOTE_VOL);
      note[i].volIndexInc = 1;
      note[i].volLevelDuration = channel[selectedChannel-1].attackVolLevelDuration;
      note[i].volLevelRemaining = channel[selectedChannel-1].attackVolLevelDuration;
      if (channel[selectedChannel-1].attackVolLevelDuration > 0) {
	note[i].volIndex = 0;
      } else {
	// optimize if there is no attack.  go directly to target volume
	note[i].volIndex = note[i].targetVolIndex;
	note[i].volumeNext = note[i].volIndex;
      }
    } else {
      if ((note[i].midiVal > NOTE_PENDING_OFF) && (note[i].midiChannel == 0)) {
	if (buttonHeld(i)) {
	  selectedChannel = i+1;
	  int currentPitchReading = map(amux.AnalogRead(PITCH_POT), 0, 1023, 36, 84);
	  if (currentPitchReading != button[i].pitchReading) {
	    // the pot has moved since the button was pressed
	    if (absolutePitch) {
	      // pitch depends on absolute position of pot
	      button[i].note = currentPitchReading;
	      button[i].pitchReading = currentPitchReading;
	    } else {
	      // pitch depends on position of pot relative to when button pressed
	      button[i].note += (currentPitchReading - button[i].pitchReading);
	      button[i].pitchReading = currentPitchReading;
	    }
	    if (button[i].note < MIDI_LOW) {
	      button[i].note = MIDI_LOW;
	    }
	    if (button[i].note > MIDI_HIGH) {
	      button[i].note = MIDI_HIGH;
	    }
	    note[i].midiVal = button[i].note;
	  }
	  if (button[i].note == button[i].lastNote) {
	    ledState = HIGH;
	  }
	} else {
	  if (note[i].envelopePhase != RELEASE) {
	    // button released
	    button[i].lastNote = note[i].midiVal;
	    note[i].envelopePhase = RELEASE;
	    note[i].targetVolIndex = 0;
	    note[i].volIndexInc = -1;
	    note[i].volLevelDuration = channel[selectedChannel-1].releaseVolLevelDuration;
	    note[i].volLevelRemaining = channel[selectedChannel-1].releaseVolLevelDuration;
	    if (channel[selectedChannel-1].releaseVolLevelDuration == 0) {
	      // optimize for immediate release
	      note[i].volIndex = 0;
	      note[i].volumeNext = 0;
	    }
	  }
	}
      }
    }
  }
}


void setPhaseIncrement(byte i) {
  // phaseInc is the number of steps we should progress through the
  // wavetable at each execution of the ISR.
  float phaseIncFloat = (note[i].frequency * N_WAVEFORM_SAMPLES) / playbackSampleRate;
  // convert the phase increment to an int
  note[i].phaseInc = (int)phaseIncFloat;
  // but also use a counter to keep track of the fractional part
  phaseIncFloat -= note[i].phaseInc;
  note[i].phaseFractionInc = (byte)(phaseIncFloat * 256.0);
}


void processEnvelope(byte i) {
  if (note[i].envelopePhase == SUSTAIN) {
    return;
  }
  if (note[i].volLevelRemaining <= 0) {
    // it is time to change the volume to the next level
    if (note[i].volIndex == note[i].targetVolIndex) {
      // this was the last volume level to use for this phase.
      // switch to new phase of envelope.
      byte phase = note[i].envelopePhase;
      if (phase == ATTACK) {
        note[i].envelopePhase = DECAY;
        // decay to the sustain volume which is a percentage of the original velocity
	byte midiChannel = note[i].midiChannel;
	if (midiChannel == 0) {
	  midiChannel = i+1; // button note case
	}
        note[i].targetVolIndex = (byte)(channel[midiChannel-1].sustainVolLevel * note[i].targetVolIndex);
        note[i].volIndexInc = -1; // decreasing volume
        note[i].volLevelDuration = channel[midiChannel-1].decayVolLevelDuration;
        note[i].volLevelRemaining = channel[midiChannel-1].decayVolLevelDuration;
      } 
      else {
        if (phase == DECAY) {
          note[i].envelopePhase = SUSTAIN;
        } 
        else {
          if (phase == RELEASE) {
            // done with this note.
            note[i].midiVal = NOTE_PENDING_OFF;
          }
        }
      }
    } 
    else {
      // not yet at the target volume.
      // Take one step closer by changing the volume.
      note[i].volIndex += note[i].volIndexInc;
      note[i].volLevelRemaining = note[i].volLevelDuration;
      note[i].volumeNext = note[i].volIndex;
    }
  }
}

void initNote(byte i, byte midiNote) {
  note[i].midiChannel = 0;
  note[i].volume = 0;
  note[i].volumeNext = 0;
  note[i].phase = 0;
  note[i].phaseInc = 0;
  note[i].phaseFraction = 0;
  note[i].phaseFractionInc = 0;
  note[i].midiVal = midiNote;
}

// find a note in the array of MAX_NOTES available notes to use
byte findNoteIndex() {
  for(byte i=0;i<numNotes;i++) {
    // find a slot for this note
    if (note[i].midiVal <= NOTE_PENDING_OFF) {
      while (note[i].midiVal != NOTE_OFF) {
	// wait until note reaches NOTE_OFF state to avoid click
      }
      return i;
    }
  }
  // there are no notes that are not being used. Check to see if any
  // are in RELEASE envelope phase and use it
  for(byte i=0;i<numNotes;i++) {
    if (note[i].envelopePhase == RELEASE) {
      note[i].midiVal = NOTE_PENDING_OFF;
      while (note[i].midiVal != NOTE_OFF) {
	// wait until note reaches NOTE_OFF state to avoid click
      }
      return i;
    }
  }
  // All notes are being used. Choose the note that is "oldest".
  byte noteIndex = 0;
  for(byte i=1;i<numNotes;i++) {
    if (note[i].startTime < note[noteIndex].startTime) {
      noteIndex = i;
    }
  }
  note[noteIndex].midiVal = NOTE_PENDING_OFF;
  while (note[noteIndex].midiVal != NOTE_OFF) {
    // wait until note reaches NOTE_OFF state to avoid click
  }
  return noteIndex;
}

void handleNoteOn(byte channelNum, byte midiNote, byte velocity) { 
#ifdef DEBUG_ENABLE
  if (debug) {
    softSerial.print("ON ");
    softSerial.print("ch ");
    softSerial.print(channelNum);
    softSerial.print(" : ");
    softSerial.print(midiNote);
    softSerial.print(" : ");
    softSerial.println(velocity);
  }
#endif
  if (channelNum > N_MIDI_CHANNELS) return;
  if ((midiNote < MIDI_LOW) || (midiNote > MIDI_HIGH)) {
    return;
  }
  if (velocity > 0) {
    byte i = findNoteIndex();
    if (velocity > MAX_VELOCITY) {
      velocity = MAX_VELOCITY;
    }
    initNote(i, midiNote);
    note[i].startTime = millis();
    note[i].midiChannel = channelNum;
    note[i].waveformBuf = waveformBuffers[channel[channelNum-1].waveform];
    note[i].envelopePhase = ATTACK;
    note[i].targetVolIndex = (byte)(((float)(velocity / 100.0)) * MAX_NOTE_VOL);
    note[i].volIndexInc = 1;
    note[i].volLevelDuration = channel[channelNum-1].attackVolLevelDuration;
    note[i].volLevelRemaining = channel[channelNum-1].attackVolLevelDuration;
    if (channel[channelNum-1].attackVolLevelDuration > 0) {
      note[i].volIndex = 0;
    } else {
      // optimize if there is no attack.  go directly to target volume
      note[i].volIndex = note[i].targetVolIndex;
      note[i].volumeNext = note[i].volIndex;
    }
  } else {
    // note off
    doNoteOff(channelNum, midiNote);
  }
}

void handleNoteOff(byte channelNum, byte midiNote, byte velocity) { 
#ifdef DEBUG_ENABLE
  if (debug) {
    softSerial.print("OFF ");
    softSerial.print("ch ");
    softSerial.print(channelNum);
    softSerial.print(" : ");
    softSerial.print(midiNote);
    softSerial.print(" : ");
    softSerial.println(velocity);
  }
#endif
  if (channelNum > N_MIDI_CHANNELS) return;
  if ((midiNote < MIDI_LOW) || (midiNote > MIDI_HIGH)) {
    return;
  }
  doNoteOff(channelNum, midiNote);
}

void doNoteOff(byte channelNum, byte midiNote) {
  for(byte i=0;i<numNotes;i++) {
    // find the note
    if ((note[i].midiVal == midiNote) && (note[i].midiChannel == channelNum)) {
      note[i].envelopePhase = RELEASE;
      note[i].targetVolIndex = 0;
      note[i].volIndexInc = -1;
      note[i].volLevelDuration = channel[channelNum-1].releaseVolLevelDuration;
      note[i].volLevelRemaining = channel[channelNum-1].releaseVolLevelDuration;
      if (channel[channelNum-1].releaseVolLevelDuration == 0) {
	// optimize for immediate release
	note[i].volIndex = 0;
	note[i].volumeNext = 0;
      }
    }
  }
}

ISR(TIMER1_OVF_vect) {
  TCNT1 = timer1Start;
  byte last;

  Synthino.output(output);

  lfoPhase += lfoPhaseInc;
  last = lfoPhaseFraction;
  lfoPhaseFraction += lfoPhaseFractionInc;
  if (lfoPhaseFraction < last) {
    // overflow in pseudo-floating point counter
    lfoPhase++;
  }
  if (lfoPhase >= N_WAVEFORM_SAMPLES) {
    lfoPhase -= N_WAVEFORM_SAMPLES;
  }


  output = SILENCE;

  int mix;
  int sum = 0;
  note_t *n;
  for(byte i=0;i<numNotes;i++) {
    n = &note[i];
    if (n->midiVal != NOTE_OFF) {
      n->phase += n->phaseInc;
      last = n->phaseFraction;
      n->phaseFraction += n->phaseFractionInc;
      if (n->phaseFraction < last) {
        // overflow in pseudo-floating point counter
        n->phase++;
      }
      if (n->phase >= N_WAVEFORM_SAMPLES) {
        if (n->midiVal != NOTE_PENDING_OFF) {
          n->phase -= N_WAVEFORM_SAMPLES;
          // don't assign volume until we cross zero
          n->volume = n->volumeNext;
        } else {
          n->midiVal = NOTE_OFF;
          continue;
        }
      }
      n->volLevelRemaining--;
      mix = pgm_read_word(n->waveformBuf + n->phase);
      mix = adjustAmplitude(mix, n->volume);
      sum += mix;
    }
  }

  // clip
  if (sum < -2048) {
    sum = -2048;
  } 
  else {
    if (sum > 2047) {
      sum = 2047;
    }
  }

  output = sum + 2048;

#ifdef DEBUG_ENABLE
  timer1End = TCNT1;
#endif
}


/*
 * Adjust amplitude using only bitshifting.  Crude but cheap.
 */
int adjustAmplitude(int v, byte level) {
  int tmp, tmp2;
  switch(level) {
  case 32:
    return v;
    break;
  case 31:
    return v - (v >> 5);
    break;
  case 30:
    return v - (v >> 4);
    break;
  case 29:
    tmp = (v >> 4);
    return (v - tmp) - (tmp >> 1);
    break;
  case 28:
    return v - (v >> 3);
    break;
  case 27:
    tmp = v >> 3;
    return (v - tmp) - (tmp >> 2);
    break;
  case 26:
    tmp = v >> 3;
    return (v - tmp) - (tmp >> 1);
    break;
  case 25:
    tmp = v >> 2;
    return (v - tmp) + (tmp >> 3);
    break;
  case 24:
    return v - (v >> 2);
    break;
  case 23:
    tmp = v >> 2;
    return (v - tmp) - (tmp >> 3);
    break;
  case 22:
    tmp = v >> 2;
    return (v - tmp) - (tmp >> 2);
    break;
  case 21:
    tmp = v >> 1;
    tmp2 = tmp >> 2;
    return tmp + tmp2 + (tmp2 >> 2); // bad
    break;
  case 20: 
    tmp = v >> 1;
    return tmp + (tmp >> 2);
    break;
  case 19:
    tmp = v >> 1;
    tmp2 = tmp >> 3;
    return tmp + tmp2 + (tmp2 >> 2); // bad
    break;
  case 18:
    tmp = v >> 1;
    return tmp + (tmp >> 3);
    break;
  case 17:
    tmp = v >> 1;
    return tmp + (tmp >> 4);
    break;
  case 16:
    return v >> 1;
    break;
  case 15:
    tmp = v >> 1;
    return tmp - (tmp >> 4);
    break;
  case 14:
    tmp = v >> 1;
    return tmp - (tmp >> 3);
    break;
  case 13:
    tmp = v >> 1;
    tmp2 = tmp >> 3;
    return (tmp - tmp2) - (tmp2 >> 1); // bad
    break;
  case 12:
    tmp = v >> 1;
    return tmp - (tmp >> 2);
    break;
  case 11:
    tmp = v >> 1;
    tmp2 = tmp >> 2;
    return (tmp - tmp2) - (tmp2 >> 2);
    break;
  case 10:
    tmp = v >> 2;
    return tmp + (tmp >> 2);
    break;
  case 9:
    tmp = v >> 2;
    return tmp + (tmp >> 3);
    break;
  case 8:
    return v >> 2;
    break;
  case 7:
    tmp = v >> 2;
    return tmp - (tmp >> 3);
    break;
  case 6:
    tmp = v >> 3;
    return tmp + (tmp >> 1);
    break;
  case 5:
    tmp = v >> 3;
    return tmp + (tmp >> 2);
    break;
  case 4:
    return v >> 3;
  case 3:
    tmp = v >> 4;
    return tmp + (tmp >> 1);
    break;
  case 2:
    return v >> 4;
  case 1:
    return v >> 5;
  default:
    return 0;
  }
}


boolean buttonPressed(byte button) {
  return buttonPressed(button, true);
}

// Return true if the button is pressed.
boolean buttonPressed(byte button, boolean updateButtonState) {
  if (digitalRead(buttonPins[button]) == LOW) {
    // the button is currently pressed
    if ((buttonState[button] == HIGH) && (updateButtonState)) {
      // if the button was not pressed before, update the state.
      buttonState[button] = LOW;
    }
    return true;
  } else {
    // The button is currently not pressed
    if ((buttonState[button] == LOW) && (updateButtonState)) {
      // if the button was pressed before, update the state.
      buttonState[button] = HIGH;
    }
    return false;
  }
}

// Return true if the button is pressed and it is a new press (not held)
boolean buttonPressedNew(byte button) {
  if (digitalRead(buttonPins[button]) == LOW) {
    // The button is currently pressed
    if (buttonState[button] == HIGH) {
      // This is a new press.
      buttonState[button] = LOW;
      return true;
    }
    // This is not a new press.
    return false; 
  } 
  else {
    // The button is currently not pressed
    if (buttonState[button] == LOW) {
      buttonState[button] = HIGH;
    }
    return false;
  }
}

// Return true if the button is pressed and it was pressed the last time the state was changed.
boolean buttonHeld(byte button) {
  if (digitalRead(buttonPins[button]) == LOW) {
    // the button is currently pressed
    if (buttonState[button] == HIGH) {
      return false;
    } else {
      return true;
    }
  } else {
    return false;
  }
}


#ifdef DEBUG_ENABLE
void doDebug() {
  if (debug) {
    if ((millis() - lastDebugPrint) >= 400) {
      lastDebugPrint = millis();

      // Print the number of instruction cycles remaining at the end of the ISR.
      // The more work you try to do in the ISR, the lower this number will become.
      // If the number of cycles remaining reaches 0, then the ISR will take up
      // all the CPU time and the code in v will not run.

      /*
      softSerial.print("mem: ");
      softSerial.print(getMemory());
      softSerial.print(" ");

      softSerial.print(" rate: ");
      softSerial.print(playbackSampleRate);
      softSerial.print(" ");

      softSerial.print("attack: ");
      softSerial.print(attackVolLevelDuration);
      softSerial.print(", decay: ");
      softSerial.print(decayVolLevelDuration);
      softSerial.print(", sustain: ");
      softSerial.print(sustainVolLevel);
      softSerial.print(", release: ");
      softSerial.print(releaseVolLevelDuration);
      softSerial.println("");
      for(byte i=0;i<numNotes;i++) {
        softSerial.print("midiVal[");
        softSerial.print(i);
        softSerial.print("]=");
        softSerial.print(note[i].midiVal);
        softSerial.print("  ");

        softSerial.print("f[");
        softSerial.print(i);
        softSerial.print("]=");
        softSerial.print(note[i].frequency);

        softSerial.print(" pi[");
        softSerial.print(i);
        softSerial.print("]=");
        softSerial.print(note[i].phaseInc);

        softSerial.print(" vol[");
        softSerial.print(i);
        softSerial.print("]=");
        softSerial.print(note[i].volume);
        softSerial.print(" | ");
      }

      softSerial.print("lfoFrequency = ");
      softSerial.print(lfoFrequency, 4);
      softSerial.print("  ");
      */

      softSerial.print(" cycles = ");
      softSerial.print(UINT16_MAX - timer1End);
      softSerial.println("");
      if ((UINT16_MAX - timer1End) < 20) {
        //softSerial.println("WARNING: ISR execution time is too long. Reduce sample rate or reduce the amount of code in the ISR.");
      }
    }
  } // if (debug)
}


int getMemory() {
  int size = 200;
  byte *buf;
  while ((size > 0) && ((buf = (byte *) malloc(--size)) == NULL));
  free(buf);
  return size;
} 

#endif

